#!/usr/bin/env python
# standard python modules
import gtk
import gtk.glade
import gobject
import gnome
import gnome.ui
import sys
import os
import shutil
import re
import subprocess
import time
import nautilusburn


# local python modules
import rdiff_interface
import version
from LogHandler import LogHandler
from Dialogs import ErrorDialog, InfoDialog, MsgDialog
import Dialogs
from Copy import RecursiveCopy, RecursiveCount

class GuiHandlers:
	##########################
	## MAIN WINDOW HANDLERS ##
	##########################
	
	### HOME PAGE ###
	def on_button_home_dir_backup_clicked(widget):
		GuiHandlers().select_set("home")
		widgets['cmb_backup_type'].set_active(1) # CDRW
		widgets['notebook1'].set_current_page(1) # backup page
		for i in range(15):
			gtk.main_iteration()
		widgets['button_do_backup'].emit("clicked")
		

	### BACKUP PAGE ###
	def on_treeview_backup_sets_cursor_changed(widget):
		setid = widget.get_cursor()[0][0]
		widgets['label_backup_set_info'].set_text(rdiff_interface.backupsets[setid]['desc'])
		widgets['label_backup_set_title'].set_markup("<big><b>"+rdiff_interface.backupsets[setid]['name']+"</b></big>")
		for widget in [	'backup_ssh_user', \
						'backup_ssh_host', \
						'backup_ssh_path']:
			widgets[widget].set_text('')
		widgets['set_destination'].child.set_text('')
		if rdiff_interface.backupsets[setid].has_key('default_dest'):
			dest = rdiff_interface.backupsets[setid]['default_dest']
			if dest[:7] == "sftp://":
				p = re.compile("sftp://([^@]+)@([^/]+)(.*)")
				matches = p.match(dest)
				if len(matches.groups()) < 3:
					user = ""
					host = ""
					path = ""
				else:
					user = matches.group(1)
					host = matches.group(2)
					path = matches.group(3)
				widgets['cmb_backup_type'].set_active(2)
				widgets['backup_ssh_user'].set_text(user)
				widgets['backup_ssh_host'].set_text(host)
				widgets['backup_ssh_path'].set_text(path)
				widgets['button_do_backup'].set_sensitive(True)
			elif dest[:7] == "cdrw://":
				device = dest.replace("cdrw://", "")
				widgets['cmb_backup_type'].set_active(1)
				found = False
				for row in widgets['cmb_backup_burner'].get_model():
					if row[2] == device:
						widgets['cmb_backup_burner'].set_active_iter(row.iter)
						found = True
				if found == False:
					widgets['cmb_backup_burner'].set_active(0)
				widgets['button_do_backup'].set_sensitive(True)
			else:
				if rdiff_interface.backupsets[setid]['removable'] == "True":
					widgets['backup_removable'].set_active(True)
				else:
					widgets['backup_removable'].set_active(False)
				widgets['cmb_backup_type'].set_active(0)
				widgets['set_destination'].child.set_text(dest)
		else:
			widgets['cmb_backup_type'].set_active(0)
		widgets['backup_controls'].set_sensitive(True)
		widgets['button_delete_set'].set_sensitive(True)

	def on_button_new_set_clicked(widget):
		widgets['druid_page_start'].show()
		widgets['druid_page_finish'].show()
		widgets['notebook1'].set_current_page(1)
		widgets['window_new_set'].show()

	def on_button_delete_set_clicked(widget):
		set = rdiff_interface.backupsets[widgets['treeview_backup_sets'].get_cursor()[0][0]]['name']
		if set == "home":
			ErrorDialog("You cannot delete this set.", widgets['window_main'])
			return True
			
		dlg = gtk.MessageDialog( \
				widgets['window_main'], \
				gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING, gtk.BUTTONS_YES_NO, \
				"Are you sure you wish to delete the backup set '%s'?\nDeleting this backup set will not delete any previous backups, nor will it affect your ability to restore them." % set)
		
		dlg.connect("response", GuiHandlers().on_delete_set_dialog_response, set)
		dlg.show()
		
	def on_delete_set_dialog_response(self, widget, response, set):
		if response == gtk.RESPONSE_YES:
			rdiff_interface.DeleteSet(set)
			self.refresh_set_list()
			widgets['backup_controls'].set_sensitive(False)
			widgets['button_delete_set'].set_sensitive(False)
			widgets['label_backup_set_title'].set_text("")
			if len(rdiff_interface.backupsets) == 0:
				widgets['label_backup_set_info'].set_text("There are no remaining backup sets.\nTo create a new backup set, click 'Add'.")
			else:
				widgets['label_backup_set_info'].set_text("Select a backup set from the list on the left.")
			widget.destroy()
		else:
			widget.destroy()
			
	def on_set_destination_changed(widget):
		if len(widget.child.get_text()) > 0:
			widgets['button_do_backup'].set_sensitive(True)
		else:
			widgets['button_do_backup'].set_sensitive(False)
	
	def on_edit1_activate(widget):
		selection = widgets['treeview_backup_sets'].get_selection()
		store, row = selection.get_selected()
		if row is not None:
			setid = int(store.get_string_from_iter(row))
			set = rdiff_interface.backupsets[setid]
			if set['name'] == "home":
				ErrorDialog("You cannot edit this set.\nIt is a complete backup of your home directory.", widgets['window_main'])
				return True
			widgets['entry_new_set_name'].set_text(set['name'])
			widgets['entry_new_set_name'].set_sensitive(False)
			widgets['entry_new_set_desc'].set_text(set['desc'])
			if set.has_key('default_dest'):
				dest = set['default_dest']
				if dest[:7] == "sftp://":
					p = re.compile("sftp://([^@]+)@([^/]+)(.*)")
					matches = p.match(dest)
					if len(matches.groups()) < 3:
						user = ""
						host = ""
						path = ""
					else:
						user = matches.group(1)
						host = matches.group(2)
						path = matches.group(3)
					widgets['cmb_dst_type'].set_active(2)
					widgets['entry_ssh_user'].set_text(user)
					widgets['entry_ssh_host'].set_text(host)
					widgets['entry_ssh_path'].set_text(path)
				elif dest[:7] == "cdrw://":
					device = dest.replace("cdrw://", "")
					widgets['cmb_backup_type'].set_active(1)
					found = False
					for row in widgets['cmb_dst_burner'].get_model():
						if row[2] == device:
							widgets['cmb_dst_burner'].set_active_iter(row.iter)
							found = True
					if found == False:
						widgets['cmb_dst_burner'].set_active(0)
				else:
					widgets['cmb_dst_type'].set_active(0)
					widgets['entry_new_set_def_dest'].set_text(dest)
			else:
				widgets['cmb_dst_type'].set_active(0)
			widgets['new_set_overwrite'].set_active(True)
			for f in set['filelist_exc']:
				if f == "**":
					continue
				if os.path.isdir(f):
					druidfilelist.append([widget.render_icon(gtk.STOCK_REMOVE,\
						gtk.ICON_SIZE_MENU, "TreeView"), widget.render_icon(\
						gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU, "TreeView"), f, False])
				else:
					druidfilelist.append([widget.render_icon(gtk.STOCK_REMOVE,\
						gtk.ICON_SIZE_MENU, "TreeView"), widget.render_icon(\
						gtk.STOCK_FILE, gtk.ICON_SIZE_MENU, "TreeView"), f, False])
			for f in set['filelist_inc']:
				if os.path.isdir(f):
					druidfilelist.append([widget.render_icon(gtk.STOCK_ADD,\
						gtk.ICON_SIZE_MENU, "TreeView"), widget.render_icon(\
						gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU, "TreeView"), f, True])
				else:
					druidfilelist.append([widget.render_icon(gtk.STOCK_ADD,\
						gtk.ICON_SIZE_MENU, "TreeView"), widget.render_icon(\
						gtk.STOCK_FILE, gtk.ICON_SIZE_MENU, "TreeView"), f, True])
		if len(druidfilelist)==1:
			widgets['druidfilelist_label'].set_text("1 item")
		else:
			widgets['druidfilelist_label'].set_text("%d items"%len(druidfilelist))
		widgets['druid_page_start'].show()
		widgets['druid_page_finish'].show()
		widgets['druid1'].set_page(widgets['druid_page_2'])
		widgets['window_new_set'].set_title("Editing backup set '%s'" % set['name'])
		widgets['window_new_set'].show()





	def on_button_do_backup_clicked(widget):
		rdiff_interface.Refresh()
		if widgets['chk_show_output_log'].get_active():
			widgets['window_output_log'].show()
		widgets['progressbar1'].set_fraction(0.05)
		widgets['lbl_status'].set_text("Starting backup...")
		widgets['notebook1'].set_sensitive(False)
		widgets['menubar1'].set_sensitive(False)
		widgets['window_main'].window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
		while gtk.events_pending():
			gtk.main_iteration()
		try:
			if widgets['cmb_backup_type'].get_active() == 0: # local
				GuiHandlers().do_local_backup()
			elif widgets['cmb_backup_type'].get_active() == 1: # cdr
				GuiHandlers().do_cdr_backup()
			elif widgets['cmb_backup_type'].get_active() == 2: # ssh
				GuiHandlers().do_ssh_backup()
		finally:
			widgets['notebook1'].set_sensitive(True)
			widgets['menubar1'].set_sensitive(True)
			widgets['window_main'].window.set_cursor(None)	
	def do_cdr_backup(self):
		"""backs up to /tmp, does a mkisofs, then nautilusburns it"""
		set_id = widgets['treeview_backup_sets'].get_cursor()[0][0]
		set = rdiff_interface.backupsets[set_id]
		buf = widgets['output_log'].get_buffer()
		buf.insert(buf.get_end_iter(), "%s: Starting backup of '%s' to CD\n" % (time.ctime(), set['name']))
		# do we have a burner to use?
		if widgets['cmb_backup_burner'].get_active() == -1:
			ErrorDialog("There are no CD burners attached to this system.", widgets['window_main'])
			buf.insert(buf.get_end_iter(), "%s: Backup failed; no CD burners detected\n\n" % time.ctime())
			return True
		drive = widgets['cmb_backup_burner'].get_model()[widgets['cmb_backup_burner'].get_active()][1]
		device_name = widgets['cmb_backup_burner'].get_model()[widgets['cmb_backup_burner'].get_active()][0]
		# check that we've got read permissions on everything we're backing up
		files_ok, filecount =  GuiHandlers().check_backup_source(set)
		if not files_ok:
			return True
		# ok, do a backup to /tmp/pybackpack
		widgets['lbl_status'].set_text("Creating backup...")
		widgets['progressbar1'].set_fraction(0.25)
		
		buf.insert(buf.get_end_iter(), "%s: Creating temporary backup in /tmp\n" % time.ctime())
		gobject.idle_add(GuiHandlers().output_log_scroll)
		dest_path = "/tmp/%s.cdimage/" % version.APPPATH
		try:
			shutil.rmtree(dest_path)
		except: # it doesn't exist, so don't worry
			pass
		try:
			os.makedirs(dest_path)
		except:
			ErrorDialog("An error occurred whilst setting up the temporary CD image directory '%s'.\nPlease ensure you have write access to this location." % dest_path, widgets['window_main'])
			widgets['lbl_status'].set_text("Ready.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Backup failed; could not create CD image directory %s\n\n" % (time.ctime(), dest_path))
			return True
		stdout = LogHandler(widgets['output_log'], widgets, filecount, True)
		stderr = LogHandler(widgets['dlg_output'])
		try:
			rdiff_interface.BackupSet(set, dest_path, stdout, stderr)
		except:
			sys.stdout = sys.__stdout__
			sys.stderr = sys.__stderr__
			if len(stderr) > 0:
				widgets['dlg_label'].set_text("An error occurred whilst backing up '%s'.\nPlease review the error output below." % set['name'])
				widgets['dlg_error'].show()
				widgets['lbl_status'].set_text("Ready.")
				widgets['progressbar1'].set_fraction(0)
				buf.insert(buf.get_end_iter(), "%s: Backup failed; could not create temporary backup in %s\n\n" % (time.ctime(), dest_path))
				ErrorDialog("Backup failed: could not create temporary backup in %s" % dest_path, widgets['window_main'])
				return True
				
		# we should have the completed backup in /tmp now, let's make an ISO
		widgets['lbl_status'].set_text("Creating CD image...")
		widgets['progressbar1'].set_fraction(0.5)
		buf.insert(buf.get_end_iter(), "%s: Creating CD image\n" % time.ctime())
		stdout = open("/tmp/%s.mkisofs.stdout" % version.APPPATH, "w")
		stderr = open("/tmp/%s.mkisofs.stderr" % version.APPPATH, "w")
		args = "mkisofs -r -o /tmp/%s.iso %s" % (version.APPPATH, dest_path)
		mkisofs = subprocess.Popen(args, stdout=stdout, stderr=stderr, shell=True)
		while mkisofs.poll() is None:
			gtk.main_iteration()
		stdout.close()
		stderr.close()
		if mkisofs.returncode != 0:
			widgets['dlg_label'].set_text("An error occurred whilst creating the CD image.\nPlease review the error log below for more information.")
			widgets['dlg_output'].get_buffer().set_text(open("/tmp/%s.mkisofs.stderr" % version.APPPATH).read())
			widgets['dlg_error'].show()
			widgets['lbl_status'].set_text("Ready.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Backup failed; could not create CD image /tmp/%s.iso\n\n" % (time.ctime(), version.APPPATH))
			return True	

		# ISO is done, let's write that bad boy...
		widgets['lbl_status'].set_text("Preparing to burn CD...")
		widgets['progressbar1'].set_fraction(0.6)
		InfoDialog("Please insert a blank CD into %s" % device_name, widgets['window_main'])
		while Dialogs.last_response == None:
			gtk.main_iteration() # kill some time until the user clicks the button
		if Dialogs.last_response == gtk.RESPONSE_CANCEL:
			widgets['lbl_status'].set_text("Ready.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Backup cancelled by user\n\n" % time.ctime())
			return True #user clicked cancel
		widgets['lbl_status'].set_text("Burning CD...")
		widgets['progressbar1'].set_fraction(0.75)
		buf.insert(buf.get_end_iter(), "%s: Starting to burn CD\n" % time.ctime())
		cdburner = nautilusburn.Recorder()
		track = nautilusburn.DataTrack()
		track.filename = "/tmp/%s.iso" % version.APPPATH
		cdburner.connect('progress-changed', GuiHandlers().on_cdburn_progress_changed)
		widgets['window_progress'].set_transient_for(widgets['window_main'])
		widgets['window_progress'].show()
		burn_result = cdburner.write_tracks(drive, [track], drive.get_max_speed_write(), nautilusburn.RECORDER_WRITE_EJECT)
		if burn_result != nautilusburn.RECORDER_RESULT_FINISHED:
			GuiHandlers().on_window_progress_response(gtk.RESPONSE_OK)
			ErrorDialog("An error occurred whilst burning the CD.", widgets['window_main'])
			widgets['lbl_status'].set_text("Ready.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Backup failed; CD burn failed\n\n" % time.ctime())
			return True
		else:
			widgets['lbl_status'].set_text("Cleaning up temporary files...")
			widgets['progressbar1'].set_fraction(0.9)
			buf.insert(buf.get_end_iter(), "%s: Cleaning up temporary files\n" % time.ctime())
			try:
				shutil.rmtree(dest_path)
				for f in [	"/tmp/%s.iso" % (device, version.APPPATH), \
							"/tmp/%s.mkisofs.stdout" % version.APPPATH, \
							"/tmp/%s.mkisofs.stderr" % version.APPPATH]:
					os.unlink(f)
			except:
				pass
			widgets['lbl_status'].set_text("Backup finished.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Backup completed successfully.\n\n" % time.ctime())
			MsgDialog("Backup completed successfully.", widgets['window_main'])
			widgets['lbl_status'].set_text("Ready.")
			return False
		
			
	def do_local_backup(self):
		destination_path = widgets['set_destination'].child.get_text()
		set_id = widgets['treeview_backup_sets'].get_cursor()[0][0]
		set = rdiff_interface.backupsets[set_id]
		if len(destination_path) == 0:
			ErrorDialog("You didn't select a location to backup to.", widgets['window_main'])
			return True
		buf = widgets['output_log'].get_buffer()
		buf.insert(buf.get_end_iter(), "%s: Starting backup of '%s' to '%s'\n" % (time.ctime(), set['name'], destination_path))
		widgets['lbl_status'].set_text("Starting backup...")
		if widgets['backup_removable'].get_active():
			InfoDialog("Please connect and mount the device for '%s'" % destination_path, widgets['window_main'])
			while Dialogs.last_response == None:
				gtk.main_iteration() # kill some time until the user clicks the button
			if Dialogs.last_response == gtk.RESPONSE_CANCEL:
				widgets['lbl_status'].set_text("Ready.")
				widgets['progressbar1'].set_fraction(0)
				buf.insert(buf.get_end_iter(), "%s: Backup cancelled by user\n\n" % time.ctime())
				return True #user clicked cancel
		if not os.path.exists(destination_path):
			try:
				os.makedirs(destination_path)
			except:
				ErrorDialog("An error occurred while trying to create '%s'.\nPlease select a different backup destination." % destination_path, widgets['window_main'])
				buf.insert(buf.get_end_iter(), "%s: Backup failed; couldn't write to '%s'\n\n" % (time.ctime(), destination_ath))
				widgets['lbl_status'].set_text("Ready.")
				return True
		# is the destination OK?
		(status, reason) = rdiff_interface.CheckDestination(destination_path)
		if not status:
			if reason == "not_empty":
				dlg = gtk.MessageDialog( \
					widgets['window_main'], \
					gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, \
					"There are already files in the backup destination you have chosen, '%s'.\nIf you backup to this location, these files will be ERASED permanently.\n\nAre you sure you want to do this?" % destination_path)
				dlg.connect("response", GuiHandlers().on_overwrite_dialog_response, destination_path, GuiHandlers().on_button_do_backup_clicked)
				dlg.show()
				widgets['lbl_status'].set_text("Ready.")
				return True
			elif reason == "no_permission":
				ErrorDialog("You don't have permission to write to '%s'.\nPlease select a different backup destination." % destination_path, widgets['window_main'])
				buf.insert(buf.get_end_iter(), "%s: Backup failed; no permission to write to '%s'\n\n" % (time.ctime(), destination_path))
				widgets['lbl_status'].set_text("Ready.")
				return True
			elif reason == "is_home_dir":
				ErrorDialog("You really, really, don't want to backup straight into your home directory; this would delete everything in there.\nSelect somewhere different!", widgets['window_main'])
				buf.insert(buf.get_end_iter(), "%s: Backup failed.\n\n" % time.ctime())
				return True
		# check that we've got read permissions on everything we're backing up
		files_ok, filecount = GuiHandlers().check_backup_source(set)
		if not files_ok:
			return True
		widgets['vbox2'].set_sensitive(False)
		widgets['menu_backup'].set_sensitive(False)
		widgets['menu_restore'].set_sensitive(False)
		stdout = LogHandler(widgets['output_log'], widgets, filecount, True)
		stderr = LogHandler(widgets['dlg_output'])
		gobject.idle_add(GuiHandlers().output_log_scroll)
		buf.insert(buf.get_end_iter(), "%s: Running rdiff-backup\n" % time.ctime())
		try:
			rdiff_interface.BackupSet(set, destination_path, stdout, stderr)
			buf.insert(buf.get_end_iter(), "%s: Backup completed.\n\n" % time.ctime())
			MsgDialog("Backup completed successfully.", widgets['window_main'])
		except:
			sys.stdout = sys.__stdout__
			sys.stderr = sys.__stderr__
			if len(stderr) > 0:
				widgets['dlg_label'].set_text("An error occurred whilst backing up '%s'.\nPlease review the error output below." % set['name'])
				widgets['dlg_error'].show()
				buf.insert(buf.get_end_iter(), "%s: Backup failed; rdiff-backup failed\n\n" % time.ctime())
		widgets['lbl_status'].set_text("Ready.")
		gotit = False
		for path in widgets['set_destination'].get_model():
			if path[0] == widgets['set_destination'].child.get_text():
				gotit = True
				break
		if not gotit:
			widgets['set_destination'].get_model().append((widgets['set_destination'].child.get_text(),))
		widgets['progressbar1'].set_fraction(0)
		widgets['vbox2'].set_sensitive(True)
		widgets['menu_backup'].set_sensitive(True)
		widgets['menu_restore'].set_sensitive(True)

	def do_ssh_backup(self):
		user = widgets['backup_ssh_user'].get_text()
		host = widgets['backup_ssh_host'].get_text()
		path = widgets['backup_ssh_path'].get_text()
		destination_path = "%s@%s::%s" % (user, host, path)
		set_id = widgets['treeview_backup_sets'].get_cursor()[0][0]
		set = rdiff_interface.backupsets[set_id]
		buf = widgets['output_log'].get_buffer()
		
		error_string = ""
		if len(user) == 0:
			error_string += "\nUser"
		if len(host) == 0:
			error_string += "\nHost"
		if len(path) == 0:
			error_string += "\nPath"
		if len(error_string) > 0:
			ErrorDialog("You did not enter enough information. Please check the" + \
				" following fields:%s" % error_string, widgets['window_main'])
			return True

		if not rdiff_interface.SSHKeysDetect(user, host):
			ErrorDialog("You do not appear to have SSH keys set up on the remote" + \
				" host.\nSSH keys allow you to connect to a remote machine without "+ \
				"typing your password, and are needed to perform SSH backups in %s." % \
				version.APPNAME + "\nFor a guide to setting up SSH keys, see:\n" + \
				"http://www.google.com/", widgets['window_main'])
			return True
		
		buf.insert(buf.get_end_iter(), "%s: Starting backup of '%s' to '%s'\n" % (time.ctime(), set['name'], host))
		widgets['lbl_status'].set_text("Starting backup...")

		# check that we've got read permissions on everything we're backing up
		files_ok, filecount = GuiHandlers().check_backup_source(set)
		if not files_ok:
			return True
		stdoutput = LogHandler(widgets['output_log'], widgets, filecount, True)
		stderror = LogHandler(widgets['dlg_output'])
		gobject.idle_add(GuiHandlers().output_log_scroll)
		buf.insert(buf.get_end_iter(), "%s: Running rdiff-backup\n" % time.ctime())
		rdiff_interface.Refresh()
		try:
			rdiff_interface.BackupSet(set, destination_path, stdoutput, stderror, False)
			inifile = os.path.join(rdiff_interface.setspath, set['path'], "set.ini")
			setfile = os.path.join(path, "rdiff-backup-data", "%s.set" % version.APPPATH)
			args = "scp %s %s@%s:%s" % (inifile, user, host, setfile)
			print args
			scp = subprocess.Popen(args, shell=True)
			while scp.poll() is None:
				gtk.main_iteration()
			if scp.poll() != 0:
				exists = 0
				try:
					exists = gnomevfs.exists("sftp://%s@%s%s" % (user, host, setfile))
				except:
					print "wtf"
				if exists == 0:
					ErrorDialog("File backup to the remote host completed, but an error occurred whilst copying the %s backup set data file. All is not lost - you can manually copy this file from\n%s\n to\n%s (on host %s)" % (version.APPNAME, inifile, setfile, host), widgets['window_main'])
					buf.insert(buf.get_end_iter(), "%s: Copy of data file failed, please copy\n%s\nto\n%s (on remote host %s)\nmanually.\n" % (time.ctime(), inifile, setfile, host))
			buf.insert(buf.get_end_iter(), "%s: Backup completed.\n\n" % time.ctime())
			MsgDialog("Backup completed successfully.", widgets['window_main'])
		except:
			sys.stdout = sys.__stdout__
			sys.stderr = sys.__stderr__
			if len(stderror) > 0:
				widgets['dlg_label'].set_text("An error occurred whilst backing up '%s'.\nPlease review the error output below." % set['name'])
				widgets['dlg_error'].show()
			else:
				ErrorDialog("An error occurred whilst backing up '%s'.", widgets['window_main'])
			buf.insert(buf.get_end_iter(), "%s: Backup failed; rdiff-backup failed\n\n" % time.ctime())
		widgets['lbl_status'].set_text("Ready.")
		widgets['progressbar1'].set_fraction(0)
	
	def check_backup_source(self, set):
		buf = widgets['output_log'].get_buffer()
		widgets['lbl_status'].set_text("Analysing backup source...")
		widgets['progressbar1'].set_pulse_step(0.01)
		widgets['progressbar1'].pulse()
		buf.insert(buf.get_end_iter(), "%s: Analysing backup source\n" % time.ctime())
		problems = []
		filecount = 0

		for path in set['filelist_inc']:
			if not os.path.exists(path):
				problems.append((path, "Missing"))
			elif path == set['default_dest']:
				buf.insert(buf.get_end_iter(), "%s: Destination directory in backup source. Omitting.\n" % time.ctime())
			elif os.path.isdir(path):
				for dirname, dirs, files in os.walk(path):
					widgets['progressbar1'].pulse()
					for file in files:
						fullpath = os.path.join(dirname,file)
						if fullpath == set['default_dest']:
							buf.insert(buf.get_end_iter(), "%s: NOTE: Destination directory in backup set. Omitting.\n" % time.ctime())
						# We might be backing up a broken symlink, but who cares, we're not following them
						elif os.path.islink(fullpath):
							filecount += 1
						elif os.access(fullpath, os.R_OK):
							filecount += 1
						else:
							problems.append((fullpath, "File"))
					for dir in dirs:
						fullpath = os.path.join(dirname,dir)
						if os.access(fullpath, os.R_OK|os.X_OK):
							filecount += 1
						else:
							problems.append((fullpath,"Directory"))
					# Stop the GUI freezing up
					while (gtk.events_pending()):
						gtk.main_iteration()

			elif os.access(path, os.R_OK):
				filecount += 1
			else:
				problems.append((path, "File"))

		if problems != []:
			ErrorDialog("Some files in the backup set are missing or not readable by your user.\nThe offending files are listed in the output log window.", widgets['window_main'])
			widgets['lbl_status'].set_text("Ready.")
			widgets['progressbar1'].set_fraction(0)
			buf.insert(buf.get_end_iter(), "%s: Permissions/file missing problems on certain files:\n" % time.ctime())
			for path, path_type in problems:
					buf.insert(buf.get_end_iter(), "%s: %s\n" % (path_type, path))
			buf.insert(buf.get_end_iter(), "%s: Backup failed.\n\n" % time.ctime())
			widgets['window_output_log'].show()
			return (False, 0)
		
		buf.insert(buf.get_end_iter(), "%s: Check complete.\n" % time.ctime())
		return (True, filecount)

	
	def on_window_progress_response(widget, response=None):
		widgets['window_progress'].hide()
		widgets['prog_button'].set_sensitive(False)
		widgets['prog_status'].set_markup("<i>Waiting for CD...</i>")
		widgets['prog_bar'].set_fraction(0)
		widgets['prog_bar'].set_text("")
	
	def output_log_scroll(self=None):
		widgets['output_log'].scroll_to_iter(widgets['output_log'].get_buffer().get_end_iter(), 0)

	def on_backup_local_dest_button_clicked(event):
		global filechooser_target
		filechooser_target = "backup_dst"
		widgets['filechooserdialog1'].set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		widgets['filechooserdialog1'].set_title("Select a backup destination")
		widgets['filechooserdialog1'].show()
		
	#def on_chk_show_output_log_toggled(widget):
	#	if widget.get_active():
	#		widgets['window_output_log'].show()
	#	else:
	#		widgets['window_output_log'].hide()
	
	def on_treeview_backup_sets_button_press_event(widget, event):
		if event.button == 3:
			widgets['set_popup'].popup(None, None, None, event.button, event.time)

	def on_cmb_backup_type_changed(widget):
		try:
			widgets['notebook3'].set_current_page(widget.get_active())
		except:
			pass
	
	def on_button4_clicked(event):
		widgets['dlg_error'].show()

	### RESTORE PAGE ###
	def on_restore_src_button_clicked(widget):
		global filechooser_target
		filechooser_target = "restore_src"
		widgets['filechooserdialog1'].set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		widgets['filechooserdialog1'].set_title("Select a location to restore from")
		filefilter = gtk.FileFilter()
		#filefilter.add_pattern("%s.set" % version.APPPATH)
		#filefilter.add_pattern("rdiff-backup-data")
		#widgets['filechooserdialog1'].set_filter(filefilter)
		widgets['notebook1'].set_current_page(2)
		widgets['filechooserdialog1'].show()
	
	def on_restore_src_changed(widget):
		global read_only_source
		if widgets['radio_restore_ssh'].get_active():
			widgets['lbl_restore_name'].set_text("Click the 'refresh' button to read this backup set")
			widgets['lbl_restore_target'].set_text('')
			widgets['lbl_restore_desc'].set_text('')
			widgets['button_do_restore'].set_sensitive(False)
			widgets['cmb_restore_increment'].get_model().clear()
			return True
		widget.child.set_text(widget.child.get_text().replace("/rdiff-backup-data", ""))
		set = rdiff_interface.ParseRestoreSrc(widget.child.get_text())
		if set is not None:
			widgets['button_do_restore'].set_sensitive(True)
			widgets['lbl_restore_target'].set_text("This data will be restored to %s" % os.path.join(os.environ['HOME'], "restored_files", set['name']))
			widgets['lbl_restore_name'].set_text(set['name'])
			widgets['lbl_restore_desc'].set_text(set['desc'])
			widgets['cmb_restore_increment'].set_model(set['increments'])
			widgets['cmb_restore_increment'].set_active(0)
			if set.has_key('readonly'):
				#widgets['read_only_source'].show()
				read_only_source = True
			else:
				#widgets['read_only_source'].hide()
				read_only_source = False
		else:
			widgets['lbl_restore_target'].set_text('')
			widgets['lbl_restore_name'].set_text('')
			widgets['lbl_restore_desc'].set_text('')
			widgets['button_do_restore'].set_sensitive(False)
			widgets['cmb_restore_increment'].set_model(gtk.ListStore(gobject.TYPE_STRING))
			#widgets['read_only_source'].hide()
			read_only_source = False
		
	def on_button_do_restore_clicked(widget):
		restore_failed = False
		ssh_restore = widgets['radio_restore_ssh'].get_active()
		buf = widgets['output_log'].get_buffer()
		if ssh_restore:
			user = widgets['restore_ssh_user'].get_text()
			host = widgets['restore_ssh_host'].get_text()
			path = widgets['restore_src'].child.get_text()
			restore_source = "%s@%s::%s" % (user, host, path)
		else:
			restore_source = widgets['restore_src'].child.get_text()
		buf.insert(buf.get_end_iter(), "%s: Starting restore operation from '%s'\n" % (time.ctime(), restore_source))
		widgets['lbl_status'].set_text("Starting restore operation...")
		widgets['progressbar1'].set_fraction(0)
		if widgets['chk_rst_show_output_log'].get_active():
			widgets['window_output_log'].show()
		rdiff_interface.Refresh()
		destination_path = os.path.join(os.environ['HOME'], "restored_files", widgets['lbl_restore_name'].get_text())
		if not os.path.exists(destination_path):
			try:
				os.makedirs(destination_path, 0755)
			except:
				ErrorDialog("An error occurred when trying to create '%s'." % destination_path, widgets['window_main'])
				buf.insert(buf.get_end_iter(), "%s: Restore failed; could not create destination path '%s'\n\n" % (time.ctime(), destination_path))
				return True
		status, reason = rdiff_interface.CheckDestination(destination_path, False) # flag even if there's an rdiff-backup-data sub dir
		if not status:
			if reason == "no_permission":
				ErrorDialog("You don't have permission to write to '%s'.\nThis shouldn't really happen, because it's in your home directory, but nevertheless I can't put any files in there. Please change the permissions accordingly and try again." % destination_path, widgets['window_main'])
			elif reason == "not_empty":
				dlg = gtk.MessageDialog( \
					widgets['window_main'], \
					gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, \
					"There are already files in '%s'.\nIf you restore your backup to this location, these files will be ERASED permanently.\n\nAre you sure you want to do this?" % destination_path)
				dlg.connect("response", GuiHandlers().on_overwrite_dialog_response, destination_path, GuiHandlers().on_button_do_restore_clicked)
				dlg.show()
#			buf.insert(buf.get_end_iter(), "%s: Restore aborted, no files were changed.\n\n" % time.ctime())
			return True
		# if we've reached this point, it's ok to do the restore
		widgets['lbl_status'].set_text("Checking restore source...")
		widgets['progressbar1'].set_fraction(0.25)
		if read_only_source:
			global total_count
			total_count = RecursiveCount(restore_source)
			global current_count
			current_count = 0
			dest_path = "/tmp/%s.restore" % version.APPPATH
			widgets['lbl_status'].set_text("Copying from CD...")
			widgets['progressbar1'].set_fraction(0.33)
			buf.insert(buf.get_end_iter(), "%s: Copying from CD (%d files) to %s\n" % (time.ctime(), total_count, dest_path))
			try:
				shutil.rmtree(dest_path)
			except:
				pass
			try:
				os.makedirs(dest_path)
				RecursiveCopy(restore_source, dest_path, GuiHandlers().on_recursive_copy_step)
			except:
				ErrorDialog("Couldn't copy CD from '%s' to '%s'.\nPlease check there is enough disk space and try again." % (restore_source, dest_path), widgets['window_main'])
				buf.insert(buf.get_end_iter(), "%s: Restore failed; could not copy CD to temporary location.\n\n" % time.ctime())
				return True
			buf.insert(buf.get_end_iter(), "%s: CD copy finished\n" % time.ctime())
			source_path = dest_path
		widgets['lbl_status'].set_text("Restoring files...")
		widgets['progressbar1'].set_fraction(0.5)
		widgets['progressbar1'].set_text("")
		widgets['button_do_restore'].set_sensitive(False)
		widgets['menu_backup'].set_sensitive(False)
		widgets['menu_restore'].set_sensitive(False)
		stdout = LogHandler(widgets['output_log'], None, 0, True)
		stderr = LogHandler(widgets['dlg_output'])
		increment = widgets['cmb_restore_increment'].get_model()[widgets['cmb_restore_increment'].get_active()][1]
		gobject.idle_add(GuiHandlers().output_log_restore_scroll)
		buf.insert(buf.get_end_iter(), "%s: Restoring files\n" % time.ctime())
		while gtk.events_pending():
			gtk.main_iteration()
		try:
			rdiff_interface.RestoreSet(restore_source, destination_path, stdout, stderr, increment)
		except:
			sys.stdout = sys.__stdout__
			sys.stderr = sys.__stderr__
			if len(stderr) > 0:
				widgets['dlg_label'].set_text("An error occurred whilst restoring from '%s'.\nPlease review the error output below." % restore_source)
				widgets['dlg_error'].show()
				buf.insert(buf.get_end_iter(), "%s: Restore failed.\n\n" % time.ctime())
				restore_failed = True
		if read_only_source:
			widgets['lbl_status'].set_text("Cleaning up %s..." % dest_path)
			widgets['progressbar1'].set_fraction(0.75)
			buf.insert(buf.get_end_iter(), "%s: Cleaning up temporary files from %s.\n" % (time.ctime(), dest_path))
			try:
				shutil.rmtree(dest_path)
			except:
				buf.insert(buf.get_end_iter(), "%s: Cleaning up temporary files failed, please manually delete %s\n" % (time.ctime(), dest_path))
		widgets['progressbar1'].set_fraction(0)
		if not restore_failed:
			buf.insert(buf.get_end_iter(), "%s: Restore succeeded.\n\n" % time.ctime())
			MsgDialog("Restore succeeded.\nThe restored files are in 'restored_files/%s' in your home directory." % widgets['lbl_restore_name'].get_text(), widgets['window_main'])
			if not ssh_restore:
				gotit = False
				for path in widgets['restore_src'].get_model():
					if path[0] == widgets['restore_src'].child.get_text():
						gotit = True
						break
				if not gotit:
					widgets['restore_src'].get_model().append((widgets['restore_src'].child.get_text(),))
		widgets['button_do_restore'].set_sensitive(True)
		widgets['menu_backup'].set_sensitive(True)
		widgets['menu_restore'].set_sensitive(True)
		widgets['lbl_status'].set_text("Ready.")
	
	def on_radio_restore_ssh_toggled(widget):
		if widget.get_active():
			widgets['restore_ssh_user'].set_sensitive(True)
			widgets['restore_ssh_host'].set_sensitive(True)
			widgets['restore_ssh_refresh'].set_sensitive(True)
		else:
			widgets['restore_ssh_user'].set_sensitive(False)
			widgets['restore_ssh_host'].set_sensitive(False)
			widgets['restore_ssh_refresh'].set_sensitive(False)
	
	def on_restore_ssh_refresh_clicked(widget):
		global read_only_source
		user = widgets['restore_ssh_user'].get_text()
		host = widgets['restore_ssh_host'].get_text()
		path = widgets['restore_src'].child.get_text()
		set = rdiff_interface.ParseRestoreSSHSrc(user, host, path)
		if set is not None:
			widgets['button_do_restore'].set_sensitive(True)
			widgets['lbl_restore_target'].set_text("This data will be restored to %s" % os.path.join(os.environ['HOME'], "restored_files", set['name']))
			widgets['lbl_restore_name'].set_text(set['name'])
			widgets['lbl_restore_desc'].set_text(set['desc'])
			widgets['cmb_restore_increment'].set_model(set['increments'])
			widgets['cmb_restore_increment'].set_active(0)
			if set.has_key('readonly'):
				#widgets['read_only_source'].show()
				read_only_source = True
			else:
				#widgets['read_only_source'].hide()
				read_only_source = False
		else:
			ErrorDialog("An error occurred while reading the remote backup set information.\nPlease check the username, host name and path that you have entered and try again.", widgets['window_main'])
			widgets['lbl_restore_target'].set_text('')
			widgets['lbl_restore_name'].set_text('')
			widgets['lbl_restore_desc'].set_text('')
			widgets['button_do_restore'].set_sensitive(False)
			widgets['cmb_restore_increment'].get_model().clear()
		
	def output_log_restore_scroll(self=None):
		widgets['output_log'].scroll_to_iter(widgets['output_log'].get_buffer().get_end_iter(), 0)
	
	#########################
	##ERROR DIALOG HANDLERS##
	#########################
	def on_dlg_error_response(widget, response):
		GuiHandlers().dlg_error_cleanup()
		return True
	
	def dlg_error_cleanup(self):
		widgets['dlg_error'].hide()
		widgets['dlg_output'].get_buffer().set_text("")
		widgets['dlg_expander'].set_expanded(False)
		widgets['dlg_error'].set_size_request(400, 115)
		widgets['dlg_error'].resize(400, 115)
	
	def on_dlg_expander_activate(widget):
		if not widget.get_expanded():
			widgets['dlg_error'].resize(400,300)
		else:
			widgets['dlg_error'].resize(400, 115)
		

	##########################
	##     MENU HANDLERS    ##
	##########################
	def on_about1_activate(event):
		if os.path.exists(os.path.join(os.path.dirname(__file__), "pybackpack_logo.png")):
			logo = gtk.gdk.pixbuf_new_from_file(os.path.join(os.path.dirname(__file__), "pybackpack_logo.png"))
			dlg = gnome.ui.About(version.APPNAME, version.VERSION, version.COPYRIGHT, version.ABOUT, version.AUTHORS, None, None, logo)
		else:
			dlg = gnome.ui.About(version.APPNAME, version.VERSION, version.COPYRIGHT, version.ABOUT, version.AUTHORS)
		dlg.set_transient_for(widgets['window_main'])
		dlg.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
		dlg.show()
	
	def on_view_output_log_activate(event):
		widgets['window_output_log'].show()


	##########################
	##    DRUID HANDLERS    ##
	##########################
	def on_druid1_cancel(widget):
		GuiHandlers().druid_cleanup()
		
	def on_druid1_delete(widget, event):
		GuiHandlers().druid_cleanup()
		return True
		
	def druid_cleanup(self):
		"""Resets the 'new backup set' druid to a 'clean' state"""
		widgets['window_new_set'].hide()
		widgets['window_new_set'].set_title("Create New Backup Set")
		widgets['druid1'].set_page(widgets['druid_page_start'])

		for widget in [	'entry_new_set_name', \
						'entry_new_set_desc', \
						'entry_new_set_def_dest', \
						'entry_ssh_user', \
						'entry_ssh_host', \
						'entry_ssh_path']:
			widgets[widget].set_text('')
		widgets['cmb_dst_type'].set_active(0)
		widgets['new_set_overwrite'].set_active(False)
		widgets['frame_set_exists_warning'].hide()
		widgets['frame_ssh_error'].hide()
		widgets['frame_empty_set_name_error'].hide()
		widgets['frame_cd_burner_error'].hide()
		widgets['entry_new_set_name'].set_sensitive(True)
		
		widgets['filechooserwidget1'].set_current_folder(os.environ['HOME'])	
		widgets['druidfilelist_label'].set_text("0 items")
		druidfilelist.clear()
	
	def on_entry_new_set_name_changed(widget):
		widget.set_text(widget.get_text().replace("/", ""))

	def on_cmb_dst_type_changed(widget):
		try:
			widgets['notebook2'].set_current_page(widget.get_active())
		except:
			pass

	def on_druid_page_2_next(widget, event):
		"""check user input"""
		if len(widgets['entry_new_set_name'].get_text()) == 0:
			widgets['frame_empty_set_name_error'].show()
			return True
		else:
			widgets['frame_empty_set_name_error'].hide()
		if rdiff_interface.SetExists(widgets['entry_new_set_name'].get_text()) and not widgets['new_set_overwrite'].get_active():
			widgets['label_set_exists_warning'].set_text("A set with the name '%s' already exists.\nYou can either enter a new name in the text box above, or overwrite the existing set." % widgets['entry_new_set_name'].get_text())
			widgets['frame_set_exists_warning'].show()
			return True
		else:
			widgets['frame_set_exists_warning'].hide()
		# ok, the set doesn't already exist - have a crack at compiling the default destination string
		if widgets['cmb_dst_type'].get_active() == 0: # local
			widgets['druid_summary_dest'].set_text(widgets['entry_new_set_def_dest'].get_text())
		elif widgets['cmb_dst_type'].get_active() == 2: # SSH
			host = widgets['entry_ssh_host'].get_text()
			user = widgets['entry_ssh_user'].get_text()
			path = widgets['entry_ssh_path'].get_text()
			if len(host) == 0 or len(user) == 0 or len(path) == 0 or path[0] != "/":
				widgets['frame_ssh_error'].show()
				return True
			else:
				widgets['frame_ssh_error'].hide()
			widgets['druid_summary_dest'].set_text("sftp://%s@%s%s" % (user, host, path))
		elif widgets['cmb_dst_type'].get_active() == 1: # CD burner
			if widgets['cmb_dst_burner'].get_active() == -1:
				return True
			widgets['druid_summary_dest'].set_text("cdrw://%s" % widgets['cmb_dst_burner'].get_model()[widgets['cmb_dst_burner'].get_active()][2])
			widgets['chk_removable_device'].set_active(True)
		widgets['druid_page_3'].set_title("Add Files/Folders to '%s'" % widgets['entry_new_set_name'].get_text())
		widgets['druid_summary_name'].set_text(widgets['entry_new_set_name'].get_text())
		widgets['druid_summary_desc'].set_text(widgets['entry_new_set_desc'].get_text())
	
	def on_druid_page_3_next(widget, event):
		if len(druidfilelist) == 0:
			ErrorDialog("You cannot create an empty backup set!", widgets['window_new_set'])
			return True

	def on_button_ssh_path_clicked(widget):
		host = widgets['entry_ssh_host'].get_text()
		user = widgets['entry_ssh_user'].get_text()
		if len(host) == 0 or len(user) == 0:
			return True
		global filechooser_target
		filechooser_target = "new_set_ssh_path"
		widgets['filechooserdialog1'].set_title("Select a default remote location to backup to")
		widgets['filechooserdialog1'].set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		widgets['filechooserdialog1'].show()
		widgets['filechooserdialog1'].set_current_folder_uri("sftp://%s@%s/" %(user, host))

	def on_button_default_dest_clicked(widget):
		global filechooser_target
		filechooser_target = "new_set_default_dest"
		widgets['filechooserdialog1'].set_title("Select a default location to backup to")
		widgets['filechooserdialog1'].set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		widgets['filechooserdialog1'].show()
	
	def on_button_refresh_cd_burners_clicked(widget):
		GuiHandlers().find_cd_burners(widgets)

	def on_druid_page_finish_prepare(widget, event):
		event.set_buttons_sensitive(False, False, False, False)
				
	def on_druid_page_finish_finish(event, something):
		filelist = []
		for f in druidfilelist:
			filelist.append((f[2], f[3]))
		try:
			rdiff_interface.WriteSet(widgets['druid_summary_name'].get_text(), \
									 widgets['druid_summary_desc'].get_text(), \
									 filelist, \
									 widgets['new_set_overwrite'].get_active(), \
									 {'default_dest': widgets['druid_summary_dest'].get_text(), \
									  'removable': widgets['chk_removable_device'].get_active()})
			GuiHandlers().refresh_set_list()
			GuiHandlers().select_set(widgets['druid_summary_name'].get_text())
			GuiHandlers().druid_cleanup()
		except OSError, e:
			widgets['druid_page_finish'].set_text("You're all done! Congratulations.\n\nERROR: Couldn't write new set %s: %s" % (widgets['entry_new_set_name'].get_text(), e))

	def on_button_add_to_set_clicked(widget):
		for f in widgets['filechooserwidget1'].get_filenames():
			if os.path.isdir(f):
				druidfilelist.append([widget.render_icon(gtk.STOCK_ADD, gtk.ICON_SIZE_MENU, "TreeView"), 
                                      widget.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU, "TreeView"), 
                                      f, True])
			else:
				druidfilelist.append([widget.render_icon(gtk.STOCK_ADD, gtk.ICON_SIZE_MENU, "TreeView"),
                                      widget.render_icon(gtk.STOCK_FILE, gtk.ICON_SIZE_MENU, "TreeView"),
                                      f, True])
		if len(druidfilelist) == 1:
			widgets['druidfilelist_label'].set_text("1 item")
		else:
			widgets['druidfilelist_label'].set_text("%d items"%len(druidfilelist))
		widgets['filechooserwidget1'].unselect_all()
	
	def on_button_exc_from_set_clicked(widget):
		for f in widgets['filechooserwidget1'].get_filenames():
			if os.path.isdir(f):
				druidfilelist.append([widget.render_icon(gtk.STOCK_REMOVE, gtk.ICON_SIZE_MENU, "TreeView"), 
                                      widget.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU, "TreeView"),
                                      f, False])
			else:
				druidfilelist.append([widget.render_icon(gtk.STOCK_REMOVE, gtk.ICON_SIZE_MENU, "TreeView"), 
                                      widget.render_icon(gtk.STOCK_FILE, gtk.ICON_SIZE_MENU, "TreeView"), 
                                      f, False])
		if len(druidfilelist) == 1:
			widgets['druidfilelist_label'].set_text("1 item")
		else:
			widgets['druidfilelist_label'].set_text("%d items"%len(druidfilelist))
		widgets['filechooserwidget1'].unselect_all()
	
	def on_button_remove_from_set_clicked(widget):
		selection = widgets['treeview_druidfilelist'].get_selection()
		store, row = selection.get_selected()
		if row is not None:
			store.remove(row)
			if len(druidfilelist)==1:
				widgets['druidfilelist_label'].set_text("1 item")
			else:
				widgets['druidfilelist_label'].set_text("%d items"%len(druidfilelist))
	
	###########################
	## FILE CHOOSER HANDLERS ##
	###########################
	def on_filechooserdialog1_show(widget):
		widget.set_current_folder(os.environ['HOME'])
		
	def on_filechooser_open_clicked(event):
		if filechooser_target == "backup_dst":
			widgets['set_destination'].child.set_text(widgets['filechooserdialog1'].get_filename())
		elif filechooser_target == "restore_src":
			widgets['restore_src'].child.set_text("")
			uri = widgets['filechooserdialog1'].get_uri()
			if uri[:7] == "sftp://":
				p = re.compile("sftp://([^@]+)@([^/]+)(.*)")
				matches = p.match(uri)
				if len(matches.groups()) < 3:
					user = ""
					host = ""
					path = ""
				else:
					user = matches.group(1)
					host = matches.group(2)
					path = matches.group(3)
				widgets['radio_restore_ssh'].set_active(True)
				widgets['restore_src'].child.set_text(path)
				widgets['restore_ssh_user'].set_text(user)
				widgets['restore_ssh_host'].set_text(host)
			elif uri[:7] == "file://":
				widgets['radio_restore_local'].set_active(True)
				widgets['restore_src'].child.set_text(uri[7:])
				widgets['restore_ssh_user'].set_text("username")
				widgets['restore_ssh_host'].set_text("host")
			else:
				ErrorDialog("I don't know what to do with this type of location!\nOnly sftp:// and file:// locations are supported.", widgets['filechooserdialog1'])
				return True
		elif filechooser_target == "restore_dst":
			widgets['restore_dst'].set_text(widgets['filechooserdialog1'].get_filename())
		elif filechooser_target == "new_set_default_dest":
			widgets['entry_new_set_def_dest'].set_text(widgets['filechooserdialog1'].get_filename())
		elif filechooser_target == "new_set_ssh_path":
			path = widgets['filechooserdialog1'].get_uri()
			host = widgets['entry_ssh_host'].get_text().replace(".", "\.")
			user = widgets['entry_ssh_user'].get_text()
			patt = re.compile("sftp:/[/]+%s@%s(.*)" % (user, host))
			match = patt.match(path)
			if len(match.groups()) > 0:
				widgets['entry_ssh_path'].set_text(match.group(1))
		widgets['filechooserdialog1'].hide()
		
	def on_filechooser_cancel_clicked(event):
		widgets['filechooserdialog1'].hide()
	
	##########################
	##  OUTPUT LOG HANDLERS ##
	##########################
	def on_button_output_log_close_clicked(widget):
		widgets['window_output_log'].hide()
	
	def on_window_output_log_delete_event(widget, event):
		widget.hide()
		return True

	##########################
	##     MISC HANDLERS    ##
	##########################
	def on_cdburn_progress_changed(self, recorder, fract, somethingelse):
		if fract > 0:
			widgets['prog_status'].set_markup("<i>Burning CD...</i>")
			widgets['prog_bar'].set_fraction(fract)
			widgets['prog_bar'].set_text("%d%%" % (fract*100))
		else:
			widgets['prog_status'].set_markup("<i>Waiting for CD...</i>")
			widgets['prog_bar'].set_fraction(0)
			widgets['prog_bar'].set_text("0%")

		if fract == 1:
			widgets['prog_button'].set_sensitive(True)
			widgets['prog_status'].set_markup("<i>Finished!</i>")
		gtk.main_iteration()
	
	def on_recursive_copy_step(self, file):
		global current_count
		global total_count
		current_count += 1.0
		widgets['progressbar1'].set_fraction(current_count/total_count)
		widgets['progressbar1'].set_text("%d%%" % int((current_count/total_count)*100))
		widgets['lbl_status'].set_text("Copying from CD - current file: %s" % os.path.basename(file))
		gtk.main_iteration()
	
	def on_overwrite_dialog_response(self, widget, response, path, func):
		if response == gtk.RESPONSE_YES:
			# user is totally sure they want to delete, so let's do it
			widget.destroy()
			shutil.rmtree(path)
			func() # go back to where we came from
		else:
			buf = widgets['output_log'].get_buffer()
			buf.insert(buf.get_end_iter(), "%s: Restore aborted, no files were changed.\n\n" % time.ctime())
			widget.destroy()
		
	def find_cd_burners(self, widgets):
		"""Uses nautilusburn to detect CD burners and populate comboboxes"""
		cdrw_list = gtk.ListStore(gobject.TYPE_STRING, object, gobject.TYPE_STRING)
		drvmon = nautilusburn.DriveMonitor()
		for drive in drvmon.get_recorder_drives():
			cdrw_list.append((drive.get_name_for_display(), drive, drive.get_device()))
		widgets['cmb_backup_burner'].set_model(cdrw_list)
		widgets['cmb_dst_burner'].set_model(cdrw_list)
		if len(cdrw_list) == 0:
			error_string = "No CD burners were detected on your system."
			widgets['lbl_cd_burner'].set_text(error_string)
			widgets['lbl_backup_burner'].set_text(error_string)
			widgets['cmb_backup_burner'].set_sensitive(False)
			widgets['cmb_dst_burner'].set_sensitive(False)
		else:
			msg_string = "Select which CD burner to use:"
			widgets['lbl_cd_burner'].set_text(msg_string)
			widgets['lbl_backup_burner'].set_text(msg_string)
			widgets['cmb_backup_burner'].set_active(0)
			widgets['cmb_dst_burner'].set_active(0)
			widgets['cmb_backup_burner'].set_sensitive(True)
			widgets['cmb_dst_burner'].set_sensitive(True)
			
	def refresh_set_list(self, _widgets=None):
		if _widgets is None:
			_widgets = widgets
		rdiff_interface.FindSets()
		setlist = _widgets['treeview_backup_sets'].get_model()
		setlist.clear()
		for set in rdiff_interface.backupsets:
			if set.has_key("default_dest"):
				if set['default_dest'][:7] == "sftp://":
					set_type = gtk.STOCK_NETWORK
				elif set['default_dest'][:7] == "cdrw://":
					set_type = gtk.STOCK_CDROM
				else:
					set_type = gtk.STOCK_DIRECTORY
			else:
				set_type = gtk.STOCK_DIRECTORY
			setlist.append([_widgets['treeview_backup_sets'].render_icon(set_type, gtk.ICON_SIZE_MENU, "TreeView"), set['name']])
		
	def select_set(self, setname):
		"""iterates through the set list and places the sets treeview cursor
				on the given set"""
		treeview = widgets['treeview_backup_sets']
		setlist = treeview.get_model()
		for row in setlist:
			if row[1] == setname:
				treeview.set_cursor(setlist.get_string_from_iter(row.iter))
				break
		
	def gtk_main_quit(event):
		# write the backup and restore MRUs back to disk
		for file, store in [("backup_mru", widgets['set_destination'].get_model()),\
			("restore_mru", widgets['restore_src'].get_model())]:
			outfile = open(os.path.join(os.environ['HOME'], ".%s" % version.\
					APPPATH, file), "w")
			for line in store:
				outfile.write("%s\n" % line[0])
			outfile.close()

		gtk.main_quit()



class BackupTool:
	def __init__(self):
		gnome.program_init(version.APPNAME, version.VERSION)
		try:
			self.widgets = gtk.glade.XML("%s/gui.glade" % os.path.dirname(__file__))
		except RuntimeError, e:
			dlg = gtk.MessageDialog( \
				None, \
				gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, \
				"Unable to start %s:\n%s" % (version.APPNAME, e))
			dlg.connect("response", lambda w: gtk.main_quit())
			sys.exit(1)

		# connect signals to handler functions
		self.widgets.signal_autoconnect(GuiHandlers.__dict__)
		
		# create a backup set called 'home' that
		# contains the user's entire home directory
		rdiff_interface.FindSets()
		exists = False
		for s in rdiff_interface.backupsets:
			if s['name'] == 'home':
				exists = True
				break
		if exists == False:
			rdiff_interface.WriteSet("home", "A complete backup of your home directory.", [(os.environ['HOME'], True)], True, {'default_dest': 'cdrw://'})
		
		# populate the backup set list on the main window
		self['treeview_backup_sets'].set_model(gtk.ListStore(gtk.gdk.Pixbuf, gobject.TYPE_STRING))
		GuiHandlers().refresh_set_list(self)
		column = gtk.TreeViewColumn()
		column.set_title('Backup Sets')
		column.set_spacing(3)
		renderer = gtk.CellRendererPixbuf()
		column.pack_start(renderer, expand=False)
		column.add_attribute(renderer, 'pixbuf', 0)
		renderer = gtk.CellRendererText()
		column.pack_start(renderer, expand=True)
		column.add_attribute(renderer, 'text', 1)
		self['treeview_backup_sets'].append_column(column)
		
		if len(self['treeview_backup_sets'].get_model()) == 0:
			self['label_backup_set_info'].set_text("To begin, you need to create a new backup set. Click the 'New' button to get started.")

		# set up druid file list
		global druidfilelist
		# list layout is: Include/Exclude icon, file/folder icon, path, include(True)/exclude(False)
		druidfilelist = gtk.ListStore(gtk.gdk.Pixbuf, gtk.gdk.Pixbuf, gobject.TYPE_STRING, bool)
		self['treeview_druidfilelist'].set_model(druidfilelist)
		self['druid_summary_filelist'].set_model(druidfilelist)
		column = gtk.TreeViewColumn()
		column.set_title("Path")
		column.set_spacing(3)
		renderer = gtk.CellRendererPixbuf()
		column.pack_start(renderer, expand=False)
		column.add_attribute(renderer, 'pixbuf', 0)
		renderer = gtk.CellRendererPixbuf()
		column.pack_start(renderer, expand=False)
		column.add_attribute(renderer, 'pixbuf', 1)
		renderer = gtk.CellRendererText()
		column.pack_start(renderer, expand=True)
		column.add_attribute(renderer, 'text', 2)
		self['treeview_druidfilelist'].append_column(column)
				
		column = gtk.TreeViewColumn()
		column.set_title("Path")
		column.set_spacing(3)
		renderer = gtk.CellRendererPixbuf()
		column.pack_start(renderer, expand=False)
		column.add_attribute(renderer, 'pixbuf', 0)
		renderer = gtk.CellRendererPixbuf()
		column.pack_start(renderer, expand=False)
		column.add_attribute(renderer, 'pixbuf', 1)
		renderer = gtk.CellRendererText()
		column.pack_start(renderer, expand=True)
		column.add_attribute(renderer, 'text', 2)
		self['druid_summary_filelist'].append_column(column)

		
		
		# other widget initialisation
		self['window_main'].set_title(version.TITLE)
		self['filechooserwidget1'].set_current_folder(os.environ['HOME'])
		self['lbl_restore_help'].set_text(self['lbl_restore_help'].get_text().\
			replace(r'%s', version.APPPATH))
		self['lbl_homebackup'].set_text(self['lbl_homebackup'].get_text().replace(\
			r'%s', os.environ['HOME']))
		self['cmb_dst_type'].set_active(0)
		self['cmb_backup_type'].set_active(0)
		self['notebook2'].set_current_page(0)
		self['notebook3'].set_current_page(0)

		# populate the list of recently used backup/restore locations
		self['set_destination'].set_model(gtk.ListStore(gobject.TYPE_STRING))
		self['restore_src'].set_model(gtk.ListStore(gobject.TYPE_STRING))
		for file, store in [("backup_mru", self['set_destination'].get_model()),\
			("restore_mru", self['restore_src'].get_model())]:
			try:
				for line in open(os.path.join(os.environ['HOME'], ".%s" % version.\
					APPPATH, file)).readlines():
					store.append((line.strip(),))
			except IOError, e:
					pass #if the file doesn't exist, don't panic!
		
		# detect what CD burners are available
		GuiHandlers().find_cd_burners(self)
		
	def __getitem__(self, key):
		return self.widgets.get_widget(key)
		
	def __getattr__(self, key):
		return self.widgets.get_widget(key)

def StartUp():
	global widgets
	widgets = BackupTool()
	gtk.main()


if __name__ == "__main__":
	StartUp()
