#!/usr/bin/env python
import rdiff_backup.Main
import os
import sys
import shutil
import gnomevfs
import gtk
import gobject
import time

import version
from LogHandler import LogHandler
from ConfigParser import SafeConfigParser, NoOptionError

base = os.environ['HOME']
setspath = "%s/.%s/sets/" % (base, version.APPPATH)

def Refresh():
	reload(rdiff_backup.Main)

def FindSets():
	"""Finds all backup sets in the user's home directory"""
	global backupsets
	backupsets = []
	if os.path.exists(setspath):
		sets = [x for x in os.listdir(setspath) if os.path.isdir(setspath+x)]
		sets.sort()
		for set in sets:
			setopts = {'path': set}
			if not os.path.exists(os.path.join(setspath, set, "filelist")):
				print "No 'filelist' for %s" % set
				continue
			setopts['filelist_inc'] = []
			setopts['filelist_exc'] = []
			filelist = open(os.path.join(setspath, set, "filelist")).readlines()
			for l in filelist:
				if l[0] == "#":
					continue
				elif l[0] == "-":
					setopts['filelist_exc'].append(l[2:].strip())
				elif l[0] == "+":
					setopts['filelist_inc'].append(l[2:].strip())
				else:
					continue
			cp = SafeConfigParser()
			if cp.read(setspath+set+"/set.ini") == []:
				print "Couldn't read in description file: %s" % setspath+set+"/set.ini"
				continue
			try:
				setopts['name'] = cp.get('Set', 'name')
				setopts['desc'] = cp.get('Set', 'desc')
				for key, val in cp.items('Set'):
					setopts[key] = val
				backupsets.append(setopts)
			except NoOptionError, e:
				print "Couldn't parse %s fully: %s" % (setspath+set+"/set.ini", e)

def SSHKeysDetect(user, host):
	"""Determines if SSH keys are setup for the remote host specified"""
	try:
		gnomevfs.open_directory("sftp://%s@%s" % (user, host))
	except:
		return False
	return True

def CheckDestination(path, ignore_rdiff_data=True):
	"""Checks if the path is writeable, and if it's got something in it"""
	if not os.access(path, os.R_OK|os.W_OK|os.X_OK):
		return (False, 'no_permission')
	if os.path.abspath(path) == os.path.abspath(os.environ['HOME']):
		return (False, 'is_home_dir')
	if len(os.listdir(path)) != 0:
		if os.listdir(path).count('rdiff-backup-data') == 1 and ignore_rdiff_data:
			return (True, '')
		else:
			return (False, 'not_empty')
	# OK to write to this path
	return (True, '')

def ParseRestoreSrc(path):
	if len(path) == 0:
		return None
	try:
		contents = os.listdir(path)
	except OSError, e:
		return None
	if contents == []:
		return None
	if contents.count('%s.set' % version.APPPATH) == 1:
		cp = SafeConfigParser()
		cp.read(os.path.join(path, '%s.set' % version.APPPATH))
		ret = {}
		for key, val in cp.items('Set'):
			ret[key] = val
		ret['increments'] = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
		log = LogHandler()
		sys.stdout = log
		args = ['--list-increments', '--parsable-output', '--terminal-verbosity', '0', path.replace("rdiff-backup-data", "")]
		Refresh()
		try:
			Run_rdiff(args)
		except OSError, e:
			"""we've got problems.
				rdiff-backup likes to be able to write to the restore source for some
				bizarre reason. If it can't, it throws a bit of a strop and crashes out.
				The most likely candidate for making this happen is a CD-ROM based backup,
				one of the more useful sources.
				To get around this (for listing increments, at least), we'll have a crack
				at getting the increments our damn self.
				We'll also set a flag in the return dict so the restore routine knows
				what it's dealing with."""
			sys.stdout = sys.__stdout__
			for f in os.listdir(path):
				if f[:15] == "current_mirror." or f[:11] == "increments.":
					rawtime = f.replace("current_mirror.", "").replace("increments.", "").replace(".data", "")
					try:
						ret['increments'].insert_before(ret['increments'][0].iter, (rawtime, rawtime))
					except IndexError, e:
						ret['increments'].append((rawtime, rawtime))
			ret['readonly'] = True
			return ret
		sys.stdout = sys.__stdout__
		for inc in str(log).splitlines():
			timestamp = inc.split()[0]
			try:
				ret['increments'].insert_before(ret['increments'][0].iter, (time.ctime(float(timestamp)), timestamp))
			except IndexError, e:
				ret['increments'].append((time.ctime(float(timestamp)), timestamp))
		return ret
	elif contents.count('rdiff-backup-data') == 1:
		return ParseRestoreSrc(os.path.join(path, 'rdiff-backup-data'))
	else:
		return None
	"""don't really need to recurse all the way down the tree :)
	else:
		for d in [os.path.join(path, x) for x in contents if os.path.isdir(os.path.join(path, x))]:
			ret = ParseRestoreSrc(d)
			if ret == None:
				continue
			else:
				return ret"""
	return None

def ParseRestoreSSHSrc(user, host, path):
	path = path.replace("/rdiff-backup-data", "")
	if len(path) == 0:
		return None
	try:
		setfile = gnomevfs.read_entire_file(os.path.join(path, "rdiff-backup-data", "%s.set" % version.APPPATH))
	except:
		return None
	open("/tmp/%s.ssh.set" % version.APPPATH, "w").write(setfile)
	cp = SafeConfigParser()
	cp.read('/tmp/%s.ssh.set' % version.APPPATH)
	ret = {}
	for key, val in cp.items('Set'):
		ret[key] = val
	ret['increments'] = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
	log = LogHandler()
	sys.stdout = log
	sshpath = "%s@%s::%s" % (user, host, path)
	args = ['--list-increments', '--parsable-output', '--terminal-verbosity', '0', sshpath]
	Refresh()
	try:
		Run_rdiff(args)
	except OSError, e:
		return None
	sys.stdout = sys.__stdout__
	for inc in str(log).splitlines():
		timestamp = inc.split()[0]
		try:
			ret['increments'].insert_before(ret['increments'][0].iter, (time.ctime(float(timestamp)), timestamp))
		except IndexError, e:
			ret['increments'].append((time.ctime(float(timestamp)), timestamp))
	return ret
	
def SetExists(name):
	return os.path.exists(setspath+name)

def ListSets():
	"""Prints a list of all backup sets"""
	for name,extra in backupsets.items():
		print "name: %s\ntitle: %s\ndescription: %s\ndefault destination: %s\n======\n\n" % (name, extra['info'][0], extra['info'][1], extra['default_dest'])

def WriteSet(name, desc, filelist, overwrite=False, extrakeys=None):
	"""Writes set.ini and filelist to a new or existing set directory"""
	try:
		os.makedirs(setspath+name)
	except OSError, e:
		if overwrite and str(e)[7:9]=="17":
			pass
		else:
			raise OSError(e)
	
	cp = SafeConfigParser()
	cp.add_section("Set")
	cp.set("Set", "name", str(name))
	cp.set("Set", "desc", str(desc))
	if extrakeys is not None:
		for key,val in extrakeys.items():
			cp.set("Set", key, str(val))
	inifile = open(setspath+name+"/set.ini", "w")
	cp.write(inifile)
	inifile.close()
	
	flist = open(setspath+name+"/filelist", "w")
	includefiles = ""
	for path, include in filelist:
		if include:
			includefiles += "+ %s\n" % path # add these to the end of the file
		else:
			flist.write("- %s\n" % path)
	flist.write(includefiles)
	flist.write("- **\n")
	flist.close()
	
def DeleteSet(name):
	if not SetExists(name):
		print "This set ('%s') doesn't exist!" % name
		return
	try:
		os.remove(setspath+name+"/set.ini")
		os.remove(setspath+name+"/filelist")
		os.rmdir(setspath+name)
	except OSError, e:
		print e
	
def SysExit(status=0):
	sys.__stderr__.write("SysExit called with status %d\n" % status)
	raise "sys.exit called"

def BackupSet(set, dest, output=sys.__stdout__, err_output=sys.__stderr__,\
	copy_set_file=True):
	"""Performs a backup of the specified set to the path dest"""
	if not SetExists(set['path']):
		print "Backup set %s not found." % set
		return
	exitfunction = sys.exit
	sys.exit = SysExit
	arglist = ['--terminal-verbosity', '5', '--verbosity', '9', '--exclude', dest, '--include-globbing-filelist', setspath+set['path']+"/filelist", "/", dest]
	sys.stdout = output
	sys.stderr = err_output
	Run_rdiff(arglist)
	if copy_set_file:
		shutil.copyfile(os.path.join(setspath, set['path'], "set.ini"), os.path.join(dest, "rdiff-backup-data", "%s.set" % version.APPPATH))
	sys.stderr = sys.__stderr__
	sys.stdout = sys.__stdout__
	sys.exit = exitfunction


def Run_rdiff(arglist):
	print >>sys.__stdout__, arglist
	rdiff_backup.Main.parse_cmdlineoptions(arglist)
	if rdiff_backup.Main.Globals.version > "0.13":
		rdiff_backup.Main.check_action()
	if rdiff_backup.Main.Globals.version < "0.13":
		rdiff_backup.Main.set_action()
	cmdpairs = rdiff_backup.Main.SetConnections.get_cmd_pairs(rdiff_backup.Main.args, rdiff_backup.Main.remote_schema, rdiff_backup.Main.remote_cmd)
	rdiff_backup.Main.Security.initialize(rdiff_backup.Main.action or "mirror", cmdpairs)
	rps = map(rdiff_backup.Main.SetConnections.cmdpair2rp, cmdpairs)
	if rdiff_backup.Main.Globals.version > "0.13":
		rdiff_backup.Main.final_set_action(rps)
	rdiff_backup.Main.misc_setup(rps)
	rdiff_backup.Main.take_action(rps)
	rdiff_backup.Main.cleanup()
	

def RestoreSet(src_path, dst_path, output=sys.__stdout__, err_output=sys.__stderr__, increment="now"):
	"""Restores the rdiff-backup backup from src to dst_path"""
	exitfunction = sys.exit
	sys.exit = SysExit
	arglist = ['--force', '--terminal-verbosity', '5', '--verbosity', '9', '-r', increment, src_path, dst_path]
	sys.stdout = output
	sys.stderr = err_output
	try:
		Run_rdiff(arglist)
	except Exception, e:
		sys.stderr.write("Error: %s\nrdiff-backup version: %s\n%s version: %s\nPlease e-mail the contents of this text box to andy@andrewprice.me.uk" % (e, rdiff_backup.Main.Globals.version, version.APPNAME, version.VERSION))
		raise AssertionError(e)
	logfile = sys.stdout
	sys.stdout = sys.__stdout__
	sys.stderr = sys.__stderr__
	sys.exit = exitfunction
